"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const change_case_1 = require("change-case");
const graphql_1 = require("graphql");
const generator_1 = require("./generator");
const context_1 = require("./context");
const document_1 = require("./document");
const graphql_2 = require("./graphql");
const special_types_1 = require("./special-types");
const language_1 = require("./language");
function printFile({ operations, fragments, path }, ast, options) {
    if (operations.length === 0 && fragments.length === 0) {
        return '';
    }
    const generator = new generator_1.default();
    const context = new context_1.default(ast, options);
    generator.printOnNewline('// This file was generated and should not be edited.');
    generator.printOnNewline('/* tslint:disable */');
    generator.printEmptyLine();
    generator.printOnNewline("import {DocumentNode} from 'graphql';");
    generator.printEmptyLine();
    const subGenerator = new generator_1.default();
    for (const operation of operations) {
        printOperation(operation, subGenerator, context);
    }
    for (const fragment of fragments) {
        printFragment(fragment, subGenerator, context);
    }
    Array.from(context.importsUsed).forEach(([importPath, specifiers]) => {
        if (importPath === path) {
            return;
        }
        let relativePath = path_1.relative(path_1.dirname(path), importPath);
        if (!relativePath.startsWith('.')) {
            relativePath = `./${relativePath}`;
        }
        language_1.printImport({
            specifiers: Array.from(specifiers).map((specifier) => `${specifier}Data`),
            source: relativePath,
        }, generator);
    });
    generator.printEmptyLine();
    Array.from(context.specialTypesUsed).reverse().forEach((type) => {
        type.print(generator);
    });
    const typesUsed = Array.from(context.typesUsed).reverse();
    if (typesUsed.length > 0) {
        language_1.printExport(() => {
            language_1.printNamespace('Schema', () => {
                Array.from(context.typesUsed).reverse().forEach((type) => {
                    language_1.printExport(() => graphql_2.printRootGraphQLType(type, generator), generator);
                    generator.printEmptyLine();
                });
            }, generator);
        }, generator);
        generator.printEmptyLine();
    }
    generator.printOnNewline(subGenerator.output);
    generator.printEmptyLine();
    generator.printOnNewline('declare const document: DocumentNode;');
    generator.printOnNewline('export default document;');
    generator.printEmptyLine();
    generator.printOnNewline('/* tslint:enable */');
    generator.printNewline();
    return generator.output;
}
exports.printFile = printFile;
function printFieldObject(fieldObject, generator, context) {
    const { field, compositeType } = fieldObject;
    const { fields = [], fragmentSpreads = [], inlineFragments = [], type } = field;
    const spreads = fragmentSpreads.map((spread) => {
        const fragment = context.ast.fragments[spread];
        let fragmentName = `${fragment.fragmentName}FragmentData`;
        if (fragment.typeCondition !== compositeType) {
            context.addUsedSpecialType(special_types_1.NullableFragment);
            fragmentName = `${special_types_1.NullableFragment.typeName}<${fragmentName}>`;
        }
        return fragmentName;
    });
    language_1.printExport(() => {
        language_1.printInterface({ name: fieldObject.name, extend: spreads }, () => {
            const seenFields = new Set();
            if (context.options.addTypename) {
                printTypenameProperty(type, '__typename', generator);
            }
            for (const field of fields) {
                const { fieldName, responseName } = field;
                if (fieldName === '__typename') {
                    if (responseName !== fieldName || !context.options.addTypename) {
                        printTypenameProperty(type, responseName, generator);
                    }
                    continue;
                }
                seenFields.add(responseName);
                printField(field, false, generator, context);
            }
            ;
            inlineFragments.forEach((fragment) => {
                fragment.fields.forEach((field) => {
                    if (seenFields.has(field.responseName)) {
                        return;
                    }
                    // If it's not on `fields`, there is no guaranteed match, so we need
                    // to say that it can be nullable
                    printField(field, true, generator, context);
                });
            });
        }, generator);
    }, generator);
}
function printTypenameProperty(type, key, generator) {
    language_1.printPropertyKey(key, false, generator);
    if (Array.isArray(type)) {
        const types = type.map((aType) => `'${getRootGraphQLType(aType).name}'`);
        generator.print(`${types.join(' | ')},`);
    }
    else {
        const finalType = getRootGraphQLType(type);
        generator.print(`'${finalType.name}',`);
    }
}
function printFragment(fragment, generator, context) {
    const { fields, fragmentName, fragmentSpreads, fragmentsReferenced, inlineFragments = [], typeCondition, possibleTypes, } = fragment;
    const typeConditionFieldMap = typeCondition.getFields();
    const finalFragmentSreads = fragmentSpreads.map((spread) => {
        const fragment = context.ast.fragments[spread];
        if (fragment.typeCondition === typeCondition) {
            return `${spread}FragmentData`;
        }
        else {
            context.addUsedSpecialType(special_types_1.NullableFragment);
            return `${special_types_1.NullableFragment.typeName}<${spread}FragmentData>`;
        }
    });
    const document = new document_1.default(`${fragmentName}FragmentData`, fragment);
    context.document = document;
    context.addUsedExternalFragments(fragmentsReferenced);
    language_1.printExport(() => {
        language_1.printInterface({
            name: document.name,
            extend: finalFragmentSreads,
        }, () => {
            const seenFields = new Set();
            if (context.options.addTypename) {
                printTypenameProperty(possibleTypes, '__typename', generator);
            }
            for (const field of fields) {
                const { fieldName, responseName } = field;
                if (fieldName === '__typename') {
                    if (responseName !== fieldName || !context.options.addTypename) {
                        printTypenameProperty(possibleTypes, responseName, generator);
                    }
                    continue;
                }
                seenFields.add(field.responseName);
                printField(field, false, generator, context);
            }
            for (const inlineFragment of inlineFragments) {
                const { fields: fragmentFields, typeCondition: fragmentTypeCondition } = inlineFragment;
                for (const field of fragmentFields) {
                    if (seenFields.has(field.responseName)) {
                        continue;
                    }
                    seenFields.add(field.responseName);
                    const forceNullable = (typeCondition !== fragmentTypeCondition) && !typeConditionFieldMap.hasOwnProperty(field.fieldName);
                    printField(field, forceNullable, generator, context);
                }
            }
        }, generator);
    }, generator);
    generator.printEmptyLine();
    if (document.fieldObjects.length) {
        language_1.printExport(() => {
            language_1.printNamespace(document.name, () => {
                for (const fieldObject of document.fieldObjects) {
                    printFieldObject(fieldObject, generator, context);
                    generator.printEmptyLine();
                }
            }, generator);
        }, generator);
        generator.printEmptyLine();
    }
}
function printOperation(operation, generator, context) {
    const { fields, variables, operationName, operationType, fragmentSpreads = [], fragmentsReferenced, } = operation;
    const document = new document_1.default(`${operationName}${change_case_1.pascal(operationType)}Data`, operation);
    context.document = document;
    context.addUsedExternalFragments(fragmentsReferenced);
    language_1.printExport(() => {
        language_1.printInterface({
            name: document.name,
            extend: fragmentSpreads.map((spread) => `${spread}FragmentData`),
        }, () => {
            fields.forEach((field) => {
                printField(field, false, generator, context);
            });
        }, generator);
    }, generator);
    generator.printEmptyLine();
    if (document.fieldObjects.length || variables.length) {
        language_1.printExport(() => {
            language_1.printNamespace(document.name, () => {
                if (variables.length) {
                    language_1.printExport(() => {
                        language_1.printInterface({
                            name: 'Variables',
                        }, () => {
                            for (const { name, type } of variables) {
                                context.addUsedType(type);
                                graphql_2.printInputGraphQLField(name, type, generator);
                            }
                        }, generator);
                    }, generator);
                    generator.printEmptyLine();
                }
                for (const fieldObject of document.fieldObjects) {
                    printFieldObject(fieldObject, generator, context);
                    generator.printEmptyLine();
                }
            }, generator);
        }, generator);
        generator.printEmptyLine();
    }
}
function printField(field, forceNullable = false, generator, context) {
    const { responseName, type } = field;
    let currentType = type;
    const nullable = forceNullable || !(currentType instanceof graphql_1.GraphQLNonNull);
    const printBefore = [];
    const printAfter = nullable ? [' | null'] : [];
    language_1.printPropertyKey(responseName, nullable, generator);
    while (currentType instanceof graphql_1.GraphQLList || currentType instanceof graphql_1.GraphQLNonNull) {
        if (currentType instanceof graphql_1.GraphQLList) {
            currentType = currentType.ofType;
            if (currentType instanceof graphql_1.GraphQLNonNull) {
                printAfter.push('[]');
                currentType = currentType.ofType;
            }
            else {
                printBefore.push('(');
                printAfter.push(' | null)[]');
            }
        }
        else {
            currentType = currentType.ofType;
        }
    }
    const finalType = currentType;
    generator.print(printBefore.join(''));
    if (finalType instanceof graphql_1.GraphQLScalarType) {
        if (graphql_2.builtInScalarMap.hasOwnProperty(finalType.name)) {
            generator.print(graphql_2.builtInScalarMap[finalType.name]);
        }
        else {
            context.addUsedType(finalType);
            generator.print(`Schema.${finalType.name}`);
        }
    }
    else if (finalType instanceof graphql_1.GraphQLEnumType) {
        context.addUsedType(finalType);
        generator.print(`Schema.${finalType.name}`);
    }
    else {
        const fieldObject = context.document.fieldObjectForField(field);
        generator.print(`${context.document.name}.${fieldObject.name}`);
    }
    generator.print(printAfter.reverse().join(''));
    generator.print(',');
}
function getRootGraphQLType(type) {
    let finalType = type;
    while (finalType instanceof graphql_1.GraphQLNonNull || finalType instanceof graphql_1.GraphQLList) {
        finalType = finalType.ofType;
    }
    return finalType;
}
